use strict;
use warnings;

package CORBA::C;

our $VERSION = '2.62';

use CORBA::C::LiteralVisitor;
use CORBA::C::NameVisitor;
use CORBA::C::LengthVisitor;
use CORBA::C::TypeVisitor;
use CORBA::C::IncludeVisitor;

1;

